@@ -43,6 +43,7 @@ from jax._src.config import (
   check_tracer_leaks as check_tracer_leaks,
   checking_leaks as checking_leaks,
   enable_custom_prng as enable_custom_prng,
+  enable_custom_vjp_by_custom_transpose as enable_custom_vjp_by_custom_transpose,
   debug_nans as debug_nans,
   debug_infs as debug_infs,
   log_compiles as log_compiles,
@@ -122,6 +123,7 @@ from jax._src.api import (
 )
 from jax.experimental.maps import soft_pmap as soft_pmap
 from jax.version import __version__ as __version__
+from jax.version import __version_info__ as __version_info__
 
 # These submodules are separate because they are in an import cycle with
 # jax and rely on the names imported above.
