@@ -11,13 +11,13 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
-
+from __future__ import annotations
 
 import collections
 from collections import namedtuple
 from contextlib import contextmanager
 import functools
-from functools import partial, partialmethod, total_ordering
+from functools import partialmethod, total_ordering
 import gc
 import itertools as it
 import operator
@@ -41,8 +41,8 @@ from jax import linear_util as lu
 
 from jax._src import source_info_util
 from jax._src.util import (safe_zip, safe_map, curry, prod, tuple_insert,
-                        tuple_delete, cache, as_hashable_function,
-                        HashableFunction)
+                        tuple_delete, as_hashable_function,
+                        HashableFunction, weakref_lru_cache)
 import jax._src.pretty_printer as pp
 
 from jax._src import traceback_util
@@ -55,13 +55,13 @@ map, unsafe_map = safe_map, map
 # -------------------- jaxprs --------------------
 
 class Jaxpr:
-  constvars: List['Var']
-  invars: List['Var']
-  outvars: List['Atom']
-  eqns: List['JaxprEqn']
+  constvars: List[Var]
+  invars: List[Var]
+  outvars: List[Atom]
+  eqns: List[JaxprEqn]
 
-  def __init__(self, constvars: Sequence['Var'], invars: Sequence['Var'],
-               outvars: Sequence['Atom'], eqns: Sequence['JaxprEqn']):
+  def __init__(self, constvars: Sequence[Var], invars: Sequence[Var],
+               outvars: Sequence[Atom], eqns: Sequence[JaxprEqn]):
     """
     Args:
       constvars: list of variables introduced for constants. Array constants are
@@ -114,7 +114,7 @@ def subjaxprs(jaxpr: Jaxpr) -> Iterator[Jaxpr]:
 
 class ClosedJaxpr:
   jaxpr: Jaxpr
-  consts: List['Any']
+  consts: List[Any]
 
   def __init__(self, jaxpr: Jaxpr, consts: Sequence):
     assert len(consts) == len(jaxpr.constvars)
@@ -160,9 +160,9 @@ def jaxpr_as_fun(closed_jaxpr: ClosedJaxpr, *args):
 
 
 class JaxprEqn(NamedTuple):
-  invars: List['Atom']
-  outvars: List['Var']
-  primitive: 'Primitive'
+  invars: List[Atom]
+  outvars: List[Var]
+  primitive: Primitive
   params: Dict[str, Any]
   source_info: source_info_util.SourceInfo
 
@@ -182,9 +182,9 @@ class Var:
   # by object id, but pretty printing might collide.
   count: int
   suffix: str
-  aval: 'AbstractValue'
+  aval: AbstractValue
 
-  def __init__(self, count: int, suffix: str, aval: 'AbstractValue'):
+  def __init__(self, count: int, suffix: str, aval: AbstractValue):
     self.count = count
     self.suffix = suffix
     self.aval = raise_to_shaped(aval)
@@ -199,6 +199,8 @@ class Var:
     return _encode_digits_alphabetic(self.count) + self.suffix
 
 def _encode_digits_alphabetic(n):
+  if n == -1:
+    return '*'
   s = ''
   while len(s) == 0 or n:
     n, i = n // 26, n % 26
@@ -211,7 +213,7 @@ def _jaxpr_vars(jaxpr):
       (v for eqn in jaxpr.eqns for v in eqn.outvars))
 
 def gensym(jaxprs: Optional[Sequence[Jaxpr]] = None,
-           suffix: str = '') -> Callable[['AbstractValue'], Var]:
+           suffix: str = '') -> Callable[[AbstractValue], Var]:
   """Produce distinct variables, printed with the optional suffix.
 
   If `jaxprs` is provided, the variables produced will be distinct from those in
@@ -230,7 +232,7 @@ def gensym(jaxprs: Optional[Sequence[Jaxpr]] = None,
 # be read. In that sense, `dropvar` is not a variable, but it is convenient to
 # treat it as a special case of one. Its `aval` is similarly inexact.
 class DropVar(Var):
-  def __init__(self, aval: 'AbstractValue'):
+  def __init__(self, aval: AbstractValue):
     super().__init__(-1, '', aval)
   def __repr__(self): return '_'
 
@@ -238,7 +240,7 @@ class Literal:
   __slots__ = ["val", "aval", "hash"]
 
   val: Any
-  aval: 'AbstractValue'
+  aval: AbstractValue
   hash: Optional[int]
 
   def __init__(self, val, aval):
@@ -356,16 +358,16 @@ def eval_jaxpr(jaxpr: Jaxpr, consts, *args):
 class Trace:
   __slots__ = ['main', 'level', 'sublevel']
 
-  main: 'MainTrace'
+  main: MainTrace
   level: int
-  sublevel: 'Sublevel'
+  sublevel: Sublevel
 
-  def __init__(self, main: 'MainTrace', sublevel: 'Sublevel') -> None:
+  def __init__(self, main: MainTrace, sublevel: Sublevel) -> None:
     self.main = main
     self.level = main.level
     self.sublevel = sublevel
 
-  def full_raise(self, val) -> 'Tracer':
+  def full_raise(self, val) -> Tracer:
     if not isinstance(val, Tracer):
       return self.pure(val)
     val._assert_live()
@@ -958,10 +960,10 @@ class AbstractValue:
     except AttributeError:
       return self.__class__.__name__
 
-  def strip_weak_type(self) -> 'AbstractValue':
+  def strip_weak_type(self) -> AbstractValue:
     return self
 
-  def strip_named_shape(self) -> 'AbstractValue':
+  def strip_named_shape(self) -> AbstractValue:
     return self
 
   def join(self, other):
@@ -1165,7 +1167,7 @@ AxisSize = Union[AxisSizeForTracing, AxisSizeForJaxprType,
 
 class DShapedArray(UnshapedArray):
   __slots__ = ['shape']
-  shape: Tuple[AxisSize, ...]  # see comment above
+  shape: Tuple[AxisSize, ...]  # noqa: F821
   array_abstraction_level = 2
 
   def __init__(self, shape, dtype, weak_type):
@@ -1677,7 +1679,7 @@ class NamedShape:
       return (self.__positional, self.__named) == (other.__positional, other.__named)
     if isinstance(other, tuple):
       return not self.__named and self.__positional == other
-    raise TypeError(f"NamedShape doesn't support comparisons with {type(other)}")
+    return False
 
   def __hash__(self):
     named = frozenset(self.__named.items())
@@ -1710,7 +1712,8 @@ class CallPrimitive(Primitive):
 
   def get_bind_params(self, params):
     new_params = dict(params)
-    subfun = lu.wrap_init(partial(eval_jaxpr, new_params.pop('call_jaxpr'), ()))
+    jaxpr = new_params.pop('call_jaxpr')
+    subfun = lu.hashable_partial(lu.wrap_init(eval_jaxpr), jaxpr, ())
     return [subfun], new_params
 
 def call_bind(primitive: CallPrimitive, fun, *args, **params):
@@ -1804,12 +1807,13 @@ class MapPrimitive(Primitive):
 
   def get_bind_params(self, params):
     new_params = dict(params)
-    subfun = lu.wrap_init(partial(eval_jaxpr, new_params.pop('call_jaxpr'), ()))
+    jaxpr = new_params.pop('call_jaxpr')
+    subfun = lu.hashable_partial(lu.wrap_init(eval_jaxpr), jaxpr, ())
     axes = new_params.pop('out_axes')
     new_params['out_axes_thunk'] = HashableFunction(lambda: axes, closure=axes)
     return [subfun], new_params
 
-def map_bind(primitive: 'MapPrimitive', fun, *args, out_axes_thunk, **params):
+def map_bind(primitive: MapPrimitive, fun, *args, out_axes_thunk, **params):
   # The new thunk depends deterministically on the old thunk and the wrapped
   # function. Any caching already has to include the wrapped function as part
   # of the key, so we only use the previous thunk for equality checks.
@@ -1876,7 +1880,7 @@ def _map_shaped_array(size: int, axis: Optional[int], aval: ShapedArray
   # TODO: Extend the named shape
   if axis is None: return aval
   return ShapedArray(tuple_delete(aval.shape, axis), aval.dtype,
-                     named_shape=aval.named_shape)
+                     named_shape=aval.named_shape, weak_type=aval.weak_type)
 
 def _unmap_shaped_array(size: int, axis_name, axis: Optional[int],
                         aval: ShapedArray) -> ShapedArray:
@@ -1885,7 +1889,7 @@ def _unmap_shaped_array(size: int, axis_name, axis: Optional[int],
   named_shape.pop(axis_name, None)
   if axis is None: return aval.replace(named_shape=named_shape)
   return ShapedArray(tuple_insert(aval.shape, axis, size), aval.dtype,
-                     named_shape=named_shape)
+                     named_shape=named_shape, weak_type=aval.weak_type)
 
 AvalMapHandlerPair = Tuple[Callable, Callable]
 aval_mapping_handlers: Dict[Type, AvalMapHandlerPair] = {
@@ -2031,7 +2035,7 @@ def do_subst_axis_names_jaxpr(jaxpr: Union[Jaxpr, ClosedJaxpr], subst: AxisSubst
     return ClosedJaxpr(new_jaxpr, consts)
   return new_jaxpr
 
-@cache()
+@weakref_lru_cache
 def used_axis_names_jaxpr(jaxpr: Union[Jaxpr, ClosedJaxpr]):
   subst = NameGatheringSubst()
   do_subst_axis_names_jaxpr(jaxpr, subst)
@@ -2130,7 +2134,7 @@ def check_jaxpr(jaxpr: Jaxpr):
     raise JaxprTypeError(msg) from None
 
 def _check_jaxpr(
-    ctx_factory: Callable[[], Tuple['JaxprPpContext', 'JaxprPpSettings']],
+    ctx_factory: Callable[[], Tuple[JaxprPpContext, JaxprPpSettings]],
     jaxpr: Jaxpr,
     in_avals: Sequence[AbstractValue]) -> None:
 
@@ -2271,7 +2275,8 @@ class JaxprPpContext:
   var_ids: DefaultDict[Var, int]
 
   def __init__(self):
-    self.var_ids = collections.defaultdict(it.count().__next__)
+    self.var_ids = collections.defaultdict(
+        it.count().__next__, {unitvar: -1})
 
 
 def pp_var(v: Var, context: JaxprPpContext) -> str:
