@@ -15,7 +15,7 @@
 from enum import IntEnum
 import numpy as np
 from collections import OrderedDict, Counter
-from typing import Callable, Sequence, Tuple, Union, Optional
+from typing import Callable, Sequence, Tuple, Union, Optional, cast, List
 import itertools as it
 from functools import partial
 
@@ -25,6 +25,7 @@ from jax import core
 from jax import linear_util as lu
 from jax import stages
 from jax._src.api import _check_callable, _check_arg
+from jax._src.config import config
 from jax._src import dispatch
 from jax._src import source_info_util
 from jax._src.lib import xla_extension_version
@@ -39,14 +40,14 @@ from jax.interpreters import pxla
 from jax.interpreters import xla
 from jax.interpreters import batching
 from jax.interpreters import partial_eval as pe
-from jax.interpreters.sharded_jit import PartitionSpec
+from jax.interpreters.pxla import PartitionSpec
 from jax._src.lib import xla_client as xc
 from jax.tree_util import (tree_map, tree_flatten, tree_unflatten,
                            treedef_is_leaf, tree_structure, treedef_tuple)
 from jax._src.tree_util import prefix_errors
 from jax._src.util import (extend_name_stack, HashableFunction, safe_zip,
                          wrap_name, wraps, distributed_debug_log,
-                         split_list, cache, tuple_insert)
+                         split_list, cache, tuple_insert, weakref_lru_cache)
 xops = xc._xla.ops
 
 class _FromGdaSingleton:
@@ -59,13 +60,17 @@ def _is_from_gda(x):
   # doesn't cause an error to avoid user confusion.
   return isinstance(x, type(FROM_GDA))
 
+_AUTOAxisResource = pxla._AUTOAxisResource
+AUTO = pxla.AUTO
+_is_auto = pxla._is_auto
+
 
 # TODO(yashkatariya): Add pjit microbenchmarks.
 def pjit(fun: Callable,
          in_axis_resources,
          out_axis_resources,
          static_argnums: Union[int, Sequence[int]] = (),
-         donate_argnums: Union[int, Sequence[int]] = ()):
+         donate_argnums: Union[int, Sequence[int]] = ()) -> stages.Wrapped:
   """Makes ``fun`` compiled and automatically partitioned across multiple devices.
 
   The returned function has semantics equivalent to those of ``fun``, but is
@@ -191,16 +196,16 @@ def pjit(fun: Callable,
     # rather than raising an error. https://github.com/google/jax/issues/2367
     in_axis_resources = tuple(in_axis_resources)
 
-  in_axis_resources, _, _ = _prepare_axis_resources(
+  in_axis_resources, _, _, in_all_auto = _prepare_axis_resources(
       in_axis_resources, "in_axis_resources")
-  out_axis_resources, _, _ = _prepare_axis_resources(
+  out_axis_resources, _, _, _ = _prepare_axis_resources(
       out_axis_resources, "out_axis_resources")
 
   static_argnums = _ensure_index_tuple(static_argnums)
   donate_argnums = _ensure_index_tuple(donate_argnums)
   donate_argnums = rebase_donate_argnums(donate_argnums, static_argnums)
 
-  def infer_params(*args, **kwargs):
+  def infer_params(*args, _global_avals=False, **kwargs):
     if kwargs:
       raise NotImplementedError("pjit does not support kwargs")
     if max(static_argnums + donate_argnums, default=-1) >= len(args):
@@ -232,20 +237,29 @@ def pjit(fun: Callable,
 
     _maybe_check_pjit_gda_mesh(args_flat, mesh)
 
+    # TODO(yashkatariya): Make sure you are not checking explicitly for `ShapedArray`.
+    # One possibility, is to only allow GDA and fully replicated inputs for AUTO.
+    if in_all_auto:
+      assert all(isinstance(a, GDA) or (isinstance(a, core.ShapedArray) and _global_avals)
+                 for a in args_flat), args_flat
+
     local_in_avals = tuple(shaped_abstractify(a) for a in args_flat)
     # TODO(yashkatariya): This is a hack. This should go away when avals have
     # is_global attribute.
-    in_positional_semantics = tuple(
-        maps._PositionalSemantics.GLOBAL if isinstance(a, GDA) else maps._positional_semantics.val
-        for a in args_flat)
-    out_positional_semantics = maps._positional_semantics.val
+    in_positional_semantics = tuple(tree_map(
+        partial(_get_in_positional_semantics, _global_avals), args_flat))
+    out_positional_semantics = (
+        maps._PositionalSemantics.GLOBAL
+        if config.jax_parallel_functions_output_gda else maps._positional_semantics.val)
 
     global_in_avals, canonicalized_in_axis_resources_flat = _process_in_axis_resources(
         mesh, local_in_avals, hashable_pytree(in_axis_resources), in_tree,
         in_positional_semantics, tuple(isinstance(a, GDA) for a in args_flat))
+
     jaxpr, canonicalized_out_axis_resources_flat = _pjit_jaxpr(
         flat_fun, mesh, global_in_avals, HashableFunction(out_tree, closure=()),
         hashable_pytree(out_axis_resources))
+
     canonicalized_in_axis_resources_flat = tree_map(
         _maybe_replace_from_gda_with_pspec,
         canonicalized_in_axis_resources_flat, tuple(args_flat))
@@ -270,9 +284,9 @@ def pjit(fun: Callable,
     out = pjit_p.bind(*args_flat, **params)
     return tree_unflatten(out_tree, out)
 
-  def lower(*args, **kwargs):
+  def lower(*args, _global_avals=False, **kwargs):
     (args_flat, flat_local_in_avals, params, in_tree, out_tree,
-     donate_argnums) = infer_params(*args, **kwargs)
+     donate_argnums) = infer_params(*args, _global_avals=_global_avals, **kwargs)
     in_is_global = _calc_is_global_sequence(
         params['in_positional_semantics'], params['in_axis_resources'])
     lowering = _pjit_lower(
@@ -283,8 +297,9 @@ def pjit(fun: Callable,
 
     args_kwargs_in_tree = treedef_tuple([in_tree, tree_flatten({})[1]])
     local_in_avals = args_kwargs_in_tree.unflatten(flat_local_in_avals)
-    return stages.Lowered(lowering, args_kwargs_in_tree, local_in_avals,
-                          out_tree, donate_argnums, no_kwargs=True)
+    return stages.Lowered.from_flat_info(
+        lowering, args_kwargs_in_tree, local_in_avals, donate_argnums, out_tree,
+        no_kwargs=True)
 
   wrapped.lower = lower
   return wrapped
@@ -369,14 +384,15 @@ def _process_in_axis_resources(mesh, local_in_avals, in_axis_resources_thunk,
   # will be raised because get_array_mapping (in local_to_global) of a
   # FROM_GDA cannot happen.
   tree_map(_check_resources_mismatch, in_axis_resources_flat, is_gda)
-  # If all inputs are either GDAs or fully replicated, then the avals are
+  # If all inputs have global semantics or fully replicated, then the avals are
   # global and the mesh should also be global. This split is because
-  # non-contiguous mesh can only be used if all inputs are either GDAs or fully
-  # replicated.
+  # non-contiguous mesh can only be used if all inputs have global semantics or
+  # fully replicated.
   # Use canonicalized in_axis_resources here because we want to treat P(None)
   # and None (for example) as equivalent.
-  if all(((not _is_from_gda(p) and p.partitions == ()) or ig)
-         for p, ig in safe_zip(canonicalized_in_axis_resources_flat, is_gda)):
+  if all(
+      (not _is_from_gda(p) and not _is_auto(p) and p.partitions == ()) or ips == maps._PositionalSemantics.GLOBAL
+      for p, ips in safe_zip(canonicalized_in_axis_resources_flat, in_positional_semantics)):
     # Shapes should be checked against non canonicalized in_axis_resources.
     # For example, partitions of () and ((),) are not equivalent, since the
     # first one is a valid spec for a scalar value, while the second is not!
@@ -535,13 +551,18 @@ def _prepare_axis_resources(axis_resources,
   # PyTrees don't treat None values as leaves, so we use an is_leaf function.
   entries, treedef = tree_flatten(axis_resources, is_leaf=lambda x: x is None)
   what = f"{arg_name} leaf specifications"
-  entries = [
+  # TODO(yashkatariya): Allow AUTO and other specification together after
+  # auto sharding pass supports user specified annotations.
+  all_auto = pxla._check_if_all_or_none_auto(entries, arg_name)
+  if not all_auto:
+    entries = [
       entry if _is_from_gda(entry) else ParsedPartitionSpec.from_user_input(
           entry, what, allow_unconstrained_dims=allow_unconstrained_dims)
       for entry in entries
-  ]
-  _check_unique_resources(entries, arg_name)
-  return tree_unflatten(treedef, entries), entries, treedef
+    ]
+    _check_unique_resources(entries, arg_name)
+  return tree_unflatten(treedef, entries), entries, treedef, all_auto
+
 
 def _check_resources_mismatch(in_axis_resources_flat, is_gda):
   if not is_gda and _is_from_gda(in_axis_resources_flat):
@@ -569,6 +590,8 @@ def _check_shapes_against_resources(what: str, is_global_shape: bool,
   for aval, aval_axis_resources in zip(flat_avals, flat_axis_resources):
     if _is_from_gda(aval_axis_resources):
       continue
+    if _is_auto(aval_axis_resources):
+      continue
     shape = aval.shape
     if len(shape) < len(aval_axis_resources):
       raise ValueError(f"One of {what} was given the resource assignment "
@@ -604,12 +627,14 @@ def _pjit_call_impl(*args, jaxpr,
   compiled = _pjit_lower(
       jaxpr, in_axis_resources, out_axis_resources,
       resource_env, donated_invars, name, in_is_global).compile()
+  if compiled._auto_spmd_lowering:
+    pxla._check_gda_xla_sharding_match(args, compiled._in_axes)
   distributed_debug_log(("Running pjit'd function", name),
                         ("mesh", resource_env.physical_mesh))
   return compiled.unsafe_call(*args)
 pjit_p.def_impl(_pjit_call_impl)
 
-@cache()
+@weakref_lru_cache
 def _pjit_lower(
     jaxpr: core.ClosedJaxpr,
     in_axis_resources: Tuple[CanonicalizedParsedPartitionSpec, ...],
@@ -787,16 +812,14 @@ def _pjit_partial_eval(trace, *in_tracers,
   if num_residuals:
     in_is_global = _calc_is_global_sequence(
         known_params['in_positional_semantics'], known_params['in_axis_resources'])
-    executable = _pjit_lower(
+    compiled = _pjit_lower(
         known_params["jaxpr"], known_params["in_axis_resources"],
         known_params["out_axis_resources"], known_params["resource_env"],
         known_params["donated_invars"], known_params["name"],
         in_is_global).compile(_allow_propagation_to_outputs=True,
                               _allow_compile_replicated=False)
-    output_op_sharding = \
-        executable.xla_executable.hlo_modules()[0].spmd_output_sharding
-    output_sharding_specs = parse_op_sharding(output_op_sharding, mesh)
-    residual_specs = tuple(output_sharding_specs[-num_residuals:])
+    _, output_ppspec = _get_ppspec_from_executable(compiled.xla_executable, mesh)
+    residual_specs = tuple(output_ppspec[-num_residuals:])
   else:
     residual_specs = ()
   known_params['out_axis_resources'] = (
@@ -934,7 +957,7 @@ pxla.custom_resource_typing_rules[pjit_p] = _resource_typing_pjit
 
 def with_sharding_constraint(x, axis_resources):
   x_flat, tree = tree_flatten(x)
-  parsed_axis_resources, entries, _ = _prepare_axis_resources(
+  parsed_axis_resources, entries, _, _ = _prepare_axis_resources(
       axis_resources, "axis_resources", allow_unconstrained_dims=True)
   axis_resources_flat = tuple(
       flatten_axes("with_sharding_constraint axis_resources",
@@ -1020,7 +1043,11 @@ pxla.custom_resource_typing_rules[sharding_constraint_p] = \
 
 # -------------------- helpers --------------------
 
-def get_array_mapping(axis_resources: ParsedPartitionSpec) -> pxla.ArrayMapping:
+def get_array_mapping(axis_resources: Union[ParsedPartitionSpec, _AUTOAxisResource]) -> pxla.ArrayMappingOrAuto:
+  # TODO(yashkatariya): Use `TypeGuard` on `_is_auto` when it is supported.
+  # Don't use `_is_auto` here to satisfy pytype and mypy.
+  if isinstance(axis_resources, _AUTOAxisResource):
+    return axis_resources
   return OrderedDict((axis, i)
                      for i, axes in enumerate(axis_resources)
                      if axes is not None for axis in axes)
@@ -1068,13 +1095,24 @@ def _calc_is_global_sequence(in_positional_semantics, in_axis_resources):
       ips == maps._PositionalSemantics.GLOBAL or p.partitions == ()
       for ips, p in safe_zip(in_positional_semantics, in_axis_resources))
 
+def _get_in_positional_semantics(global_avals: bool, arg) -> maps._PositionalSemantics:
+  if isinstance(arg, GDA):
+    return maps._PositionalSemantics.GLOBAL
+  if global_avals and isinstance(arg, core.ShapedArray):
+    return maps._PositionalSemantics.GLOBAL
+  return maps._positional_semantics.val
 
 def _create_cpspec(x):
-  return x if _is_from_gda(x) else CanonicalizedParsedPartitionSpec(x)
+  return x if _is_from_gda(x) or _is_auto(x) else CanonicalizedParsedPartitionSpec(x)
 
 def _maybe_replace_from_gda_with_pspec(
-    in_axis_resources_flat: CanonicalizedParsedPartitionSpec, arg) -> CanonicalizedParsedPartitionSpec:
+    in_axis_resources_flat: Union[CanonicalizedParsedPartitionSpec, _AUTOAxisResource],
+    arg) -> Union[CanonicalizedParsedPartitionSpec, _AUTOAxisResource]:
   if isinstance(arg, GDA):
+    # TODO(yashkatariya): Use `TypeGuard` on `_is_auto` when it is supported.
+    # Don't use `_is_auto` here to satisfy pytype and mypy.
+    if isinstance(in_axis_resources_flat, _AUTOAxisResource):
+      return in_axis_resources_flat
     gda_cpspec = CanonicalizedParsedPartitionSpec(
         ParsedPartitionSpec.from_user_input(
             arg.mesh_axes, arg_name="GDA mesh_axes"))
@@ -1195,11 +1233,15 @@ def explode_superdims(sizes, dims):
     final_dims += reversed(new_dims)
   return final_dims
 
-def parse_op_sharding(op_sharding, mesh):
+def parse_flatten_op_sharding(op_sharding: xc.OpSharding,
+                              mesh: pxla.Mesh) -> Sequence[ParsedPartitionSpec]:
   if op_sharding.type == xc.OpSharding.Type.TUPLE:
-    return [parse_op_sharding(s, mesh) for s in op_sharding.tuple_shardings]
+    out: List[ParsedPartitionSpec] = []
+    for s in op_sharding.tuple_shardings:
+      out.extend(parse_flatten_op_sharding(s, mesh))
+    return out
   elif op_sharding.type == xc.OpSharding.Type.REPLICATED:
-    return REPLICATED
+    return [REPLICATED]
   elif op_sharding.type == xc.OpSharding.Type.OTHER:
     mesh_shape = mesh.shape
     mesh_axis_order = unflatten_array(mesh.shape, op_sharding.tile_assignment_devices)
@@ -1223,6 +1265,30 @@ def parse_op_sharding(op_sharding, mesh):
         raise NotImplementedError("Unhandled OpSharding type. Please open a bug report!")
     if replicate_on_last_tile_dim:
       partitions = partitions[:-1]
-    return ParsedPartitionSpec('<internally generated spec>', partitions)
+    return [ParsedPartitionSpec('<internally generated spec>', partitions)]
   else:
     raise AssertionError("Unhandled OpSharding type. Please open a bug report!")
+
+
+def _get_partition_spec(ppspec: Sequence[ParsedPartitionSpec]) -> Sequence[PartitionSpec]:
+  return [pxla.array_mapping_to_axis_resources(cast(pxla.ArrayMapping, get_array_mapping(p)))
+          for p in ppspec]
+
+
+def _get_ppspec_from_executable(executable, mesh) -> Tuple[Sequence[ParsedPartitionSpec], Sequence[ParsedPartitionSpec]]:
+  input_op_shardings: Sequence[xc.OpSharding] = executable.hlo_modules()[0].spmd_parameters_shardings
+  output_op_sharding: xc.OpSharding = executable.hlo_modules()[0].spmd_output_sharding
+  in_ppspec: List[ParsedPartitionSpec] = []
+  for sharding in input_op_shardings:
+    in_ppspec.extend(parse_flatten_op_sharding(sharding, mesh))
+  out_ppspec = parse_flatten_op_sharding(output_op_sharding, mesh)
+  return in_ppspec, out_ppspec
+
+
+def _get_sharding_from_executable(
+    executable, mesh: pxla.Mesh
+) -> Tuple[Tuple[PartitionSpec, ...], Tuple[PartitionSpec, ...]]:
+  in_ppspec, out_ppspec = _get_ppspec_from_executable(executable, mesh)
+  out_partition_spec = _get_partition_spec(out_ppspec)
+  in_partition_spec = _get_partition_spec(in_ppspec)
+  return tuple(in_partition_spec), tuple(out_partition_spec)
