@@ -28,13 +28,13 @@ from typing import (Any, Callable, Dict, List, Optional, Sequence, Set, Tuple,
 from typing_extensions import Protocol
 import warnings
 
+import jax
 from jax import core
 from jax import linear_util as lu
 from jax._src import ad_util
 from jax._src import device_array
 from jax._src import dtypes
 from jax._src.lib.mlir import ir
-from jax._src.lib.mlir.dialects import builtin
 from jax._src.lib.mlir.dialects import chlo
 from jax._src.lib.mlir.dialects import mhlo
 from jax._src.lib.mlir.dialects import func as func_dialect
@@ -46,6 +46,15 @@ import jax.interpreters.partial_eval as pe
 import jax.interpreters.xla as xla
 import numpy as np
 
+# TODO(jakevdp): remove this when minimum_jaxlib_version >= 0.3.3
+if jax._src.lib.mlir_api_version >= 4:
+  FuncOp = func_dialect.FuncOp  # pytype: disable=module-attr
+else:
+  from jax._src.lib.mlir.dialects import builtin  # pylint: disable=import-not-at-top
+  FuncOp = builtin.FuncOp  # pytype: disable=module-attr
+# mypy gets confused by conditional imports, so alias to Any for now.
+FuncOpType = Any
+
 
 map, unsafe_map = util.safe_map, map
 zip, unsafe_zip = util.safe_zip, zip
@@ -106,7 +115,7 @@ def dtype_to_ir_type(dtype: Union[np.dtype, np.generic]) -> ir.Type:
         f"No dtype_to_ir_type handler for dtype: {dtype}") from err
   return ir_type_factory()
 
-def _array_ir_types(aval: core.ShapedArray) -> ir.Type:
+def _array_ir_types(aval: core.ShapedArray) -> Sequence[ir.Type]:
   return (ir.RankedTensorType.get(aval.shape, dtype_to_ir_type(aval.dtype)),)
 
 ir_type_handlers: Dict[Type[core.AbstractValue],
@@ -341,23 +350,26 @@ class ModuleContext:
   name_stack: NameStack
 
   # Cached primitive lowerings.
-  cached_primitive_lowerings: Dict[Any, builtin.FuncOp]
+  cached_primitive_lowerings: Dict[Any, FuncOpType]
 
   @property
   def axis_env(self) -> xla.AxisEnv:
     return self.axis_context.axis_env
 
   def __init__(
-      self, platform: str, axis_context: AxisContext, name_stack: NameStack,
+      self,
+      platform: str,
+      axis_context: AxisContext,
+      name_stack: NameStack,
       context: Optional[ir.Context] = None,
       module: Optional[ir.Module] = None,
       ip: Optional[ir.InsertionPoint] = None,
       symbol_table: Optional[ir.SymbolTable] = None,
-      cached_primitive_lowerings: Optional[Dict[Any, builtin.FuncOp]] = None):
+      cached_primitive_lowerings: Optional[Dict[Any, FuncOpType]] = None):
     assert platform is not None
     self.context = context or make_ir_context()
     self.module = module or ir.Module.create(loc=ir.Location.unknown(self.context))
-    self.ip = ip or ir.InsertionPoint(self.module.operation.opview.body)
+    self.ip = ip or ir.InsertionPoint(self.module.body)
     self.symbol_table = symbol_table or ir.SymbolTable(self.module.operation)
     self.platform = platform
     self.axis_context = axis_context
@@ -471,7 +483,7 @@ def lower_jaxpr_to_module(
     # TODO(tomhennigan): At call time we should mark these buffers as deleted.
     unused_donations = [str(a) for a, d in zip(in_avals, donated_args)
                         if d]
-    msg = "See an explanation at https://jax.readthedocs.io/en/latest/notebooks/faq.html#buffer-donation."
+    msg = "See an explanation at https://jax.readthedocs.io/en/latest/faq.html#buffer-donation."
     if platform not in platforms_with_donation:
       msg = f"Donation is not implemented for {platform}.\n{msg}"
     warnings.warn(f"Some donated buffers were not usable: {', '.join(unused_donations)}.\n{msg}")
@@ -524,14 +536,18 @@ def _set_up_aliases(avals_in, avals_out, donated_args):
   return input_output_aliases, out_donated_args
 
 def lower_jaxpr_to_fun(
-    ctx: ModuleContext, name: str, jaxpr: core.ClosedJaxpr, *,
-    public: bool = False, replace_units_with_dummy: bool = False,
+    ctx: ModuleContext,
+    name: str,
+    jaxpr: core.ClosedJaxpr,
+    *,
+    public: bool = False,
+    replace_units_with_dummy: bool = False,
     replace_tokens_with_dummy: bool = False,
     replicated_args: Optional[Sequence[bool]] = None,
     arg_shardings: Optional[Sequence[Optional[xc.OpSharding]]] = None,
     result_shardings: Optional[Sequence[Optional[xc.OpSharding]]] = None,
     input_output_aliases: Optional[Sequence[Optional[int]]] = None
-  ) -> builtin.FuncOp:
+) -> FuncOpType:
   """Lowers jaxpr and its callees to an IR function.
 
   Assumes that an MLIR context, location, and insertion point are set.
@@ -565,7 +581,7 @@ def lower_jaxpr_to_fun(
   flat_input_types = util.flatten(input_types)
   flat_output_types = util.flatten(output_types)
   ftype = ir.FunctionType.get(flat_input_types, flat_output_types)
-  func_op = builtin.FuncOp(name, ftype, ip=ctx.ip)
+  func_op = FuncOp(name, ftype, ip=ctx.ip)
   func_op.attributes["sym_visibility"] = ir.StringAttr.get(
       "public" if public else "private")
   ctx.symbol_table.insert(func_op)
@@ -647,7 +663,7 @@ def lower_jaxpr_to_fun(
   return func_op
 
 def _emit_lowering_rule_as_fun(lowering_rule,
-                               ctx: LoweringRuleContext) -> builtin.FuncOp:
+                               ctx: LoweringRuleContext) -> FuncOpType:
   """Emits the contents of a lowering rule as a private function."""
   input_types = map(aval_to_ir_types, ctx.avals_in)
   output_types = map(aval_to_ir_types, ctx.avals_out)
@@ -655,7 +671,7 @@ def _emit_lowering_rule_as_fun(lowering_rule,
   flat_output_types = util.flatten(output_types)
   ftype = ir.FunctionType.get(flat_input_types, flat_output_types)
   assert ctx.primitive is not None
-  func_op = builtin.FuncOp(ctx.primitive.name, ftype, ip=ctx.module_context.ip)
+  func_op = FuncOp(ctx.primitive.name, ftype, ip=ctx.module_context.ip)
   func_op.attributes["sym_visibility"] = ir.StringAttr.get("private")
   ctx.module_context.symbol_table.insert(func_op)
   entry_block = func_op.add_entry_block()
@@ -705,10 +721,11 @@ def jaxpr_subcomp(ctx: ModuleContext, jaxpr: core.Jaxpr,
     with source_info_util.user_context(eqn.source_info.traceback), loc:
       if eqn.primitive in _platform_specific_lowerings[ctx.platform]:
         rule = _platform_specific_lowerings[ctx.platform][eqn.primitive]
+      elif eqn.primitive in xla._backend_specific_translations[ctx.platform]:
+        rule = xla_fallback_lowering(eqn.primitive)
       elif eqn.primitive in _lowerings:
         rule = _lowerings[eqn.primitive]
-      elif (eqn.primitive in xla._translations or
-            eqn.primitive in xla._backend_specific_translations[ctx.platform]):
+      elif eqn.primitive in xla._translations:
         rule = xla_fallback_lowering(eqn.primitive)
       else:
         raise NotImplementedError(
@@ -725,7 +742,7 @@ def jaxpr_subcomp(ctx: ModuleContext, jaxpr: core.Jaxpr,
       out_nodes = tuple(map(wrap_singleton_ir_values, ans))
     except TypeError as e:
       raise ValueError("Output of translation rule must be iterable: "
-                       f"{eqn}") from e
+                       f"{eqn}, got output {ans}") from e
 
     assert all(isinstance(v, tuple) for v in out_nodes), (ans, eqn)
     assert all(isinstance(v, ir.Value) for w in out_nodes for v in w), (ans, eqn)
@@ -812,23 +829,31 @@ register_lowering(ad_util.add_jaxvals_p, add_jaxvals_lowering)
 register_lowering(ad_util.stop_gradient_p, lambda ctx, x: [x])
 
 
+def compare_mhlo(x, y, direction, type):
+  """Creates mhlo.CompareOp."""
+  if jax._src.lib.mlir_api_version >= 5:
+    return mhlo.CompareOp(x, y, mhlo.ComparisonDirectionAttr.get(direction),
+                          mhlo.ComparisonTypeAttr.get(type))
+  dims = ir.RankedTensorType(x.type).shape
+  bool_shape = ir.RankedTensorType.get(dims, ir.IntegerType.get_signless(1))
+  if jax._src.lib.mlir_api_version >= 3:
+    return mhlo.CompareOp(bool_shape, x, y,
+                          mhlo.ComparisonDirectionAttr.get(direction),
+                          mhlo.ComparisonTypeAttr.get(type))
+  return mhlo.CompareOp(bool_shape, x, y, ir.StringAttr.get(direction),
+                        ir.StringAttr.get(type))
+
 def _minmax_mhlo(op, cmp, x, y):
   """Min/max that compares complex values lexicographically as pairs."""
   tensor_type = ir.RankedTensorType(x.type)
   if ir.ComplexType.isinstance(tensor_type.element_type):
     rx = mhlo.RealOp(x).result
     ry = mhlo.RealOp(y).result
-    dims = [tensor_type.get_dim_size(i) for i in range(tensor_type.rank)]
-    bool_shape = ir.RankedTensorType.get(dims, ir.IntegerType.get_signless(1))
-    real_eq = mhlo.CompareOp(bool_shape, rx, ry, ir.StringAttr.get("EQ"),
-                             ir.StringAttr.get("FLOAT"))
-    real_cmp = mhlo.CompareOp(bool_shape, rx, ry,
-                              ir.StringAttr.get(cmp),
-                              ir.StringAttr.get("FLOAT"))
-    imag_cmp = mhlo.CompareOp(bool_shape, mhlo.ImagOp(x).result,
-                              mhlo.ImagOp(y).result,
-                              ir.StringAttr.get(cmp),
-                              ir.StringAttr.get("FLOAT"))
+    real_eq = compare_mhlo(rx, ry, "EQ", "FLOAT")
+    real_cmp = compare_mhlo(rx, ry, cmp, "FLOAT")
+    imag_cmp = compare_mhlo(
+        mhlo.ImagOp(x).result,
+        mhlo.ImagOp(y).result, cmp, "FLOAT")
     which = mhlo.SelectOp(real_eq, imag_cmp, real_cmp).result
     return mhlo.SelectOp(which, x, y)
   else:
@@ -850,9 +875,8 @@ def convert_mhlo(x, aval_in, aval_out):
       compare_type = "SIGNED"
     else:
       compare_type = "UNSIGNED"
-    return mhlo.CompareOp(
-        aval_to_ir_type(aval_out), x, full_like_aval(0, aval_in),
-        ir.StringAttr.get("NE"), ir.StringAttr.get(compare_type)).result
+    return compare_mhlo(x, full_like_aval(0, aval_in), "NE",
+                         compare_type).result
   return mhlo.ConvertOp(aval_to_ir_type(aval_out), x).result
 
 def _wrap_with_spmd_op(name: str,
@@ -936,6 +960,7 @@ def xla_fallback_lowering(prim: core.Primitive):
     submodule = ir.Module.parse(submodule_str)
     callee_name = None
     for op in submodule.body.operations:
+      op = typing.cast(FuncOpType, op)
       module_ctx.module.body.append(op)
       if op.name.value == "main":
         op.attributes["sym_name"] = ir.StringAttr.get(f"xla_fallback_{prim.name}")
@@ -954,8 +979,13 @@ def xla_fallback_lowering(prim: core.Primitive):
                                flatten_lowering_ir_args(args)).result
     if not prim.multiple_results:
       return [call]
-    flat_results = [mhlo.GetTupleElementOp(typ, call, i32_attr(i)).result
-                    for i, typ in enumerate(flat_output_types)]
+    if jax._src.lib.mlir_api_version < 6:
+      flat_results = [mhlo.GetTupleElementOp(typ, call, i32_attr(i)).result
+                      for i, typ in enumerate(flat_output_types)]
+    else:
+      flat_results = [mhlo.GetTupleElementOp(call, i32_attr(i)).result
+                      for i in range(len(flat_output_types))]
+
     return util.unflatten(flat_results, map(len, output_types))
   return fallback
 
