@@ -34,8 +34,8 @@ from jax._src.api_util import flattened_fun_in_tree, flatten_fun_nokwargs
 from jax._src.tree_util import (PyTreeDef, treedef_tuple, tree_unflatten,
                                 tree_leaves)
 from jax._src.util import (unzip2, safe_zip, safe_map, toposort, split_list,
-                           merge_lists, partition_list, cache, OrderedSet,
-                           as_hashable_function)
+                           merge_lists, partition_list, OrderedSet,
+                           as_hashable_function, weakref_lru_cache)
 from jax.core import (Trace, Tracer, Jaxpr, Literal, get_aval, AbstractValue,
                       unit, unitvar, abstract_unit, ClosedJaxpr, new_jaxpr_eqn,
                       ConcreteArray, raise_to_shaped, Var, Atom, JaxprEqn,
@@ -94,7 +94,6 @@ class PartialVal(tuple):
     """Either the stored known value, or the given 'val'."""
     known = self.get_known()
     return known if known is not None else val
-    breakpoint()
 
 
 class JaxprTrace(Trace):
@@ -743,7 +742,7 @@ def tracers_to_jaxpr(
   config.jax_enable_checks and core.check_jaxpr(jaxpr)
   return jaxpr, const_vals, env_vals
 
-@cache()
+@weakref_lru_cache
 def convert_constvars_jaxpr(jaxpr: Jaxpr) -> Jaxpr:
   """Moves the constvars to the start of invars."""
   config.jax_enable_checks and core.check_jaxpr(jaxpr)
@@ -806,7 +805,7 @@ def partial_eval_jaxpr(jaxpr: ClosedJaxpr, unknowns: Sequence[bool],
   instantiate = tuple(instantiate) if isinstance(instantiate, list) else instantiate
   return _partial_eval_jaxpr(jaxpr, tuple(unknowns), instantiate)
 
-@cache()
+@weakref_lru_cache
 def _partial_eval_jaxpr(jaxpr, unknowns, instantiate):
   f = lu.wrap_init(core.jaxpr_as_fun(jaxpr))
 
@@ -1172,7 +1171,7 @@ def _dce_jaxpr(closed_jaxpr: ClosedJaxpr, outputs: Sequence[bool], drop_outputs=
   new_jaxpr = _dce_open_jaxpr(closed_jaxpr.jaxpr, tuple(outputs), drop_outputs)
   return core.ClosedJaxpr(new_jaxpr, closed_jaxpr.consts)
 
-@cache()
+@weakref_lru_cache
 def _dce_open_jaxpr(jaxpr: Jaxpr, outputs: Tuple[bool, ...], drop_outputs=False) -> Jaxpr:
   # This dead-code elimination is pretty rudimentary, and in particular doesn't
   # nontrivially DCE through scan, call, or other higher-order primitives.
@@ -1192,7 +1191,7 @@ def _dce_open_jaxpr(jaxpr: Jaxpr, outputs: Tuple[bool, ...], drop_outputs=False)
   new_eqns = new_eqns[::-1]
   return Jaxpr(jaxpr.constvars, jaxpr.invars, new_outvars, new_eqns)
 
-@cache()
+@weakref_lru_cache
 def _drop_vars(jaxpr: Jaxpr, drop_ins: Tuple[bool, ...], drop_outs: Tuple[bool, ...]):
   return Jaxpr(jaxpr.constvars,
                [v for v, d in zip(jaxpr.invars, drop_ins) if not d],
@@ -1500,13 +1499,11 @@ class DynamicJaxprTrace(core.Trace):
     dim_tracers = _get_tracers_only_in_shapes(tracers)
     in_avals = _tracers_to_avals(dim_tracers + tracers)
     keep_inputs = [False] * len(dim_tracers) + [True] * len(tracers)
-    name_stack = source_info_util.current_name_stack()
     with core.new_sublevel():
       jaxpr, out_avals, consts = trace_to_subjaxpr_dynamic(
         f, self.main, in_avals, keep_inputs=keep_inputs)
     if params.get('inline', False):
-      with source_info_util.set_name_stack(name_stack):
-        return core.eval_jaxpr(jaxpr, consts, *dim_tracers, *tracers)
+      return core.eval_jaxpr(jaxpr, consts, *dim_tracers, *tracers)
     source_info = source_info_util.current()
     env = {v: t for v, t in zip((*jaxpr.constvars, *jaxpr.invars),
                                 (*consts, *dim_tracers, *tracers))
@@ -1627,10 +1624,12 @@ class DynamicJaxprTrace(core.Trace):
 
     transpose_flat, in_tree2 = flatten_fun_nokwargs(
         lu.wrap_init(transpose), treedef_tuple((res_tree, out_tree)))
-    transpose_jaxpr, in_avals2, transpose_consts = trace_to_subjaxpr_dynamic(
-        transpose_flat, self.main, in_avals_t)
-    closed_transpose_jaxpr = core.ClosedJaxpr(
-        convert_constvars_jaxpr(transpose_jaxpr), ())
+
+    main_ = ref(self.main)
+    # the following thunk evaluates to a pair: transpose_jaxpr, transpose_consts
+    transpose_jaxpr_thunk = _memoize(
+        lambda: trace_to_subjaxpr_dynamic(
+            transpose_flat, main_(), in_avals_t)[::2])
 
     out_tracers = [DynamicJaxprTracer(self, a) for a in out_avals]
     invars = map(self.getvar, tracers)
@@ -1638,9 +1637,9 @@ class DynamicJaxprTrace(core.Trace):
     outvars = map(self.makevar, out_tracers)
     eqn = new_jaxpr_eqn([*constvars, *invars], outvars, prim,
                         dict(call_jaxpr=closed_call_jaxpr,
-                             transpose_jaxpr=(closed_transpose_jaxpr,
-                                              transpose_consts),
-                             num_consts=len(call_consts)),
+                             transpose_jaxpr_thunk=transpose_jaxpr_thunk,
+                             out_types=out_types, res_tree=res_tree,
+                             lin_tree=lin_tree, out_tree=out_tree),
                         source_info_util.current())
     self.frame.eqns.append(eqn)
     return out_tracers
@@ -1776,6 +1775,10 @@ def trace_to_subjaxpr_dynamic(fun: lu.WrappedFun, main: core.MainTrace,
     out_tracers = map(trace.full_raise, ans)
     jaxpr, consts = frame.to_jaxpr(out_tracers)
     del fun, main, trace, frame, in_tracers, out_tracers, ans
+  if not config.jax_dynamic_shapes:
+    # TODO(frostig,mattjj): check_jaxpr is incomplete under dynamic
+    # shapes; remove this guard when it is
+    config.jax_enable_checks and core.check_jaxpr(jaxpr)
   return jaxpr, [v.aval for v in jaxpr.outvars], consts
 
 @contextlib.contextmanager
