@@ -14,7 +14,7 @@
 
 from functools import partial
 from typing import Callable, Iterable, Optional, Tuple, Union
-
+from warnings import warn
 from absl import logging
 import numpy as np
 
@@ -289,35 +289,6 @@ xla.register_translation(sharded_call_p, _sharded_jit_translation_rule)
 mlir.register_lowering(sharded_call_p, _sharded_jit_lowering)
 
 
-class _UnconstrainedPartitionSingleton:
-
-  def __str__(self):
-    return "UNCONSTRAINED"
-
-
-# Unconstrained sentinel value for PartitionSpec, representing a dimension for
-# which the user wants XLA to assign the best partitioning.
-# TODO(yashkatariya): May rename to AUTO.
-_UNCONSTRAINED_PARTITION = _UnconstrainedPartitionSingleton()
-
-
-class PartitionSpec(tuple):
-  """Tuple of integer specifying how a value should be partitioned.
-
-  Each integer corresponds to how many ways a dimension is partitioned. We
-  create a separate class for this so JAX's pytree utilities can distinguish it
-  from a tuple that should be treated as a pytree.
-  """
-  def __new__(cls, *partitions):
-    return tuple.__new__(PartitionSpec, partitions)
-
-  def __repr__(self):
-    return "PartitionSpec%s" % tuple.__repr__(self)
-
-  """A sentinel value representing a dim is unconstrained."""
-  UNCONSTRAINED = _UNCONSTRAINED_PARTITION
-
-
 def sharded_jit(
     fun: Callable,
     in_parts,
@@ -403,6 +374,9 @@ def sharded_jit(
   Returns:
     A version of ``fun`` that will be distributed across multiple devices.
   """
+  warn("`sharded_jit` is deprecated. Please use `pjit` instead. "
+       "See https://jax.readthedocs.io/en/latest/jax-101/08-pjit.html for more information.",
+       DeprecationWarning)
   if num_partitions is not None:
     nparts = num_partitions
   else:
@@ -493,7 +467,7 @@ def _sharding_constraint_lowering(ctx, x_node, partitions):
 mlir.register_lowering(sharding_constraint_p, _sharding_constraint_lowering)
 
 
-def with_sharding_constraint(x, partitions: Optional[PartitionSpec]):
+def with_sharding_constraint(x, partitions: Optional[pxla.PartitionSpec]):
   """Identity-like function that specifies how ``x`` should be sharded.
 
   WARNING: this feature is still under active development! It may not work well,
