@@ -14,12 +14,6 @@
 
 # flake8: noqa: F401
 from jax._src.ops.scatter import (
-  index as index,
-  index_add as index_add,
-  index_mul as index_mul,
-  index_update as index_update,
-  index_min as index_min,
-  index_max as index_max,
   segment_sum as segment_sum,
   segment_prod as segment_prod,
   segment_min as segment_min,
